#!/usr/bin/env python
#
# bitmaptoggle.py - A button which toggles between two bitmaps.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`BitmapToggleButton` class, a button
which toggles between two bitmaps.
"""


import wx
import wx.lib.newevent as wxevent


class BitmapToggleButton(wx.ToggleButton):
    """A :class:`wx.ToggleButton` which stores a boolean state
    (``True``/``False``), and displays one of two bitmaps depending on the
    state.

    When the user presses the button, the boolean state and the displayed
    bitmap changes accordingly. A :data:`EVT_BITMAP_TOGGLE` event is generated
    on button presses.
    """

    def __init__(self, parent, *args, **kwargs):
        """Create a ``BitmapToggleButton``.

        :arg parent:   The :mod:`wx` parent window.

        :arg trueBmp:  Required. A :class:`wx.Bitmap` to display when
                       the button state is ``True``.

        :arg falseBmp: Optional. A :class:`wx.Bitmap` to display when the
                       button state is ``False``.

        All other arguemnts are passed through to the
        ``wx.ToggleButton.__init__`` method.
        """

        trueBmp  = kwargs.pop('trueBmp',  None)
        falseBmp = kwargs.pop('falseBmp', None)
        style    = kwargs.get('style',    0)

        # wx.ToggleButton(style=BU_NOTEXT)
        # causes segfault under mac
        if wx.Platform == '__WXMAC__' and (style | wx.BU_NOTEXT):
            kwargs['style'] = style & ~wx.BU_NOTEXT

        wx.ToggleButton.__init__(self, parent, *args, **kwargs)

        self.__trueBmp  = None
        self.__falseBmp = None

        self.Bind(wx.EVT_TOGGLEBUTTON, self.__onToggle)

        self.SetBitmap(trueBmp, falseBmp)


    def SetBitmap(self, trueBmp, falseBmp=None):
        if falseBmp is None:
            falseBmp = trueBmp

        self.__trueBmp  = trueBmp
        self.__falseBmp = falseBmp

        self.__updateBitmap()


    def SetValue(self, state):
        """Sets the current boolean state of this ``BitmapToggleButton``."""
        wx.ToggleButton.SetValue(self, state)
        self.__updateBitmap()


    def __updateBitmap(self):
        """Toggles the state of this ``BitmapToggleButton``."""

        trueBmp  = self.__trueBmp
        falseBmp = self.__falseBmp

        if trueBmp is None:
            return

        if self.GetValue(): wx.ToggleButton.SetBitmapLabel(self, trueBmp)
        else:               wx.ToggleButton.SetBitmapLabel(self, falseBmp)


    def __onToggle(self, ev):
        """Called when this button is pressed.

        Flips the button state, and emits a :data:`BitmapToggleEvent`.
        """
        self.__updateBitmap()

        ev = BitmapToggleEvent(value=self.GetValue())
        ev.SetEventObject(self)

        wx.PostEvent(self, ev)


_BitmapToggleEvent, _EVT_BITMAP_TOGGLE = wxevent.NewEvent()


EVT_BITMAP_TOGGLE = _EVT_BITMAP_TOGGLE
"""Identifier for the :data:`BitmapToggleEvent` event."""


BitmapToggleEvent = _BitmapToggleEvent
"""Event emitted when a :class:`BitmapToggleButton` is pushed.

Contains one attribute, ``value``, which contains the current button state
(``True`` or ``False``).
"""



#!/usr/bin/env python
#
# floatslider.py - Floating point slider widgets.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides two classes, the :class:`FloatSlider` and
:class:`SliderSpinPanel`, which allow the user to modify floating point
values.
"""


import numpy as np

import wx
import wx.lib.newevent as wxevent

from . import floatspin
from . import numberdialog


class FloatSlider(wx.Panel):
    """Floating point slider widget.

    This class is an alternative to :class:`wx.Slider`, which supports
    floating point numbers of any range. The desired range is transformed into
    the internal range :math:`[-2^{31}, 2^{31-1}]`.

    A :class:`EVT_SLIDER` event is emitted whenever the user changes the
    value.
    """

    def __init__(self,
                 parent,
                 value=None,
                 minValue=None,
                 maxValue=None,
                 style=0):
        """Create a ``FloatSlider``.

        The following style flags are available:

         .. autosummary::
            FS_MOUSEWHEEL
            FS_INTEGER

        :arg parent:   The :mod:`wx` parent object.

        :arg value:    Initial slider value.

        :arg minValue: Minimum slider value.

        :arg maxValue: Maximum slider value.

        :arg style:    A combination of :data:`FS_MOUSEWHEEL` and
                       :data:`FS_INTEGER`.
        """

        if value    is None: value    = 0
        if minValue is None: minValue = 0
        if maxValue is None: maxValue = 1

        self.__sliderMin   = -2 ** 31
        self.__sliderMax   =  2 ** 31 - 1
        self.__sliderRange = abs(self.__sliderMax - self.__sliderMin)
        self.__integer     = style & FS_INTEGER > 0

        wx.Panel.__init__(self, parent)

        self.__slider = wx.Slider(self,
                                  minValue=self.__sliderMin,
                                  maxValue=self.__sliderMax,
                                  style=wx.SL_HORIZONTAL)
        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__slider, flag=wx.EXPAND, proportion=1)
        self.SetSizer(self.__sizer)

        if style & FS_MOUSEWHEEL:
            self.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        # Under GTK, slider widgets absorb
        # mousewheel events, so we bind our
        # own handler to prevent this, and
        # propagate the mousewheel event
        # upwards.
        elif wx.Platform == '__WXGTK__':
            def wheel(ev):
                self.GetParent().GetEventHandler().ProcessEvent(ev)
            self.Bind(wx.EVT_MOUSEWHEEL, wheel)

        self.__lastSliderValue = None
        self.__SetRange(minValue, maxValue)
        self.SetValue(value)

        # Under GTK, sliders don't seem to
        # be sized appropriately, so setting
        # the min size will force them to
        # have a good size
        if self.HasFlag(wx.VERTICAL): self.__slider.SetMinSize((-1,  150))
        else:                         self.__slider.SetMinSize((150, -1))

        # Under CentOS6/GTK, mouse clicks on
        # the slider do not update the slider
        # position - users have to click+drag
        # on the slider thumb. Workaround
        # is to re-implement mouse dragging
        # logic on GTK.
        if wx.Platform == '__WXGTK__':
            self.__dragging = False
            self.__slider.Bind(wx.EVT_LEFT_DOWN, self.__onMouseDown)
            self.__slider.Bind(wx.EVT_MOTION,    self.__onMouseMove)
            self.__slider.Bind(wx.EVT_LEFT_UP,   self.__onMouseUp)

        self.__slider.Bind(wx.EVT_SLIDER, self.__onSlider)


    def __onSlider(self, ev):
        """Called when the child ``wx.Slider`` instance generates an
        ``EVT_SLIDER`` event. If the slider value has changed, the
        event is propagated upwards.
        """

        newSliderValue = self.__slider.GetValue()

        # Under OSX, slider values are emitted on
        # both mouse down and mouse up events, so
        # we make sure to only propagate the event
        # if the slider value has actually changed.
        if self.__lastSliderValue is None or \
           not np.isclose(self.__lastSliderValue, newSliderValue):
            self.__lastSliderValue = newSliderValue
            ev.Skip()


    def __onMouseDown(self, ev):
        """Only called when running on GTK. Sets an internal dragging
        flag, and calls :meth:`__mouseMove`.
        """
        self.__dragging = True
        self.__onMouseMove(ev)


    def __onMouseMove(self, ev):
        """Only called when running on GTK. Updates the slider value based
        on the mouse location.
        """
        if not self.__dragging:
            return

        dataMin, dataMax = self.GetRange()
        width, height    = self.GetClientSize().Get()
        vertical         = self.GetWindowStyle() & wx.VERTICAL > 0
        x                = float(ev.GetX())
        y                = float(ev.GetY())

        if vertical: value = dataMin + (y / height) * (dataMax - dataMin)
        else:        value = dataMin + (x / width)  * (dataMax - dataMin)

        if self.SetValue(value):
            ev = wx.PyCommandEvent(wx.EVT_SLIDER.typeId, self.GetId())
            wx.PostEvent(self.GetEventHandler(), ev)


    def __onMouseUp(self, ev):
        """Only called when running on GTK. Clears the internal dragging
        flag.
        """
        self.__dragging = False


    def __onMouseWheel(self, ev):
        """If the :data:`FS_MOUSEWHEEL` style is set, this method is called
        when the mouse wheel is spun over the slider widget.

        Increases/decreases the slider value accordingly.
        """
        if not self.IsEnabled():
            return

        wheelDir  = ev.GetWheelRotation()
        increment = (self.__realMax - self.__realMin) / 100.0

        if   wheelDir < 0: increment = -increment
        elif wheelDir > 0: increment =  increment
        else:              return

        if self.SetValue(self.GetValue() + increment):
            ev = wx.PyCommandEvent(wx.EVT_SLIDER.typeId, self.GetId())
            wx.PostEvent(self.GetEventHandler(), ev)


    def GetRange(self):
        """Return a tuple containing the (minimum, maximum) slider values."""
        return (self.__realMin, self.__realMax)


    def __SetRange(self, minValue, maxValue):
        """Set the minimum/maximum slider values.

        This logic is not in the public :meth:`SetRange` method so
        we can overcome a chicken-and-egg problem in :meth:`__init__` -
        :meth:`SetValue` needs :attr:`__realMin` and :attr:`__realMax`
        to be set, but :meth:`SetRange` needs to retrieve the value
        before setting :attr:`__realMin` and :attr:`__realMax`.
        """

        minValue  = float(minValue)
        maxValue  = float(maxValue)

        if self.__integer:
            minValue  = int(round(minValue))
            maxValue  = int(round(maxValue))

        self.__realMin   = minValue
        self.__realMax   = maxValue
        self.__realRange = abs(self.__realMax - self.__realMin)


    def SetRange(self, minValue, maxValue):
        """Set the minimum/maximum slider values."""

        if minValue > maxValue:
            raise ValueError('Min cannot be greater than max '
                             '({} > {})'.format(minValue, maxValue))

        # wx.Slider values change when their bounds
        # are changed. It does this to keep the
        # slider position the same as before, but I
        # think it is more appropriate to keep the
        # slider value the same ...
        oldValue = self.GetValue()
        self.__SetRange(minValue, maxValue)
        self.SetValue(oldValue)


    def GetMin(self):
        """Return the minimum slider value."""
        return self.GetRange()[0]


    def GetMax(self):
        """Return the maximum slider value."""
        return self.GetRange()[1]


    def SetMin(self, minValue):
        """Set the minimum slider value."""
        self.SetRange(minValue, self.GetMax())


    def SetMax(self, maxValue):
        """Set the maximum slider value."""
        self.SetRange(self.GetMin(), maxValue)


    def __sliderToReal(self, value):
        """Converts the given value from slider space to real space."""

        if self.__realRange == 0:
            return 0

        value = self.__realMin + (value - self.__sliderMin) * \
            (float(self.__realRange) / self.__sliderRange)

        if self.__integer: return int(round(value))
        else:              return value


    def __realToSlider(self, value):
        """Converts the given value from real space to slider space."""

        if self.__integer:
            value = int(round(value))

        if self.__realRange == 0:
            return 0

        value = self.__sliderMin + (value - self.__realMin) * \
            (self.__sliderRange / float(self.__realRange))
        return int(round(value))


    def SetValue(self, value):
        """Set the slider value. Returns ``True`` if the value changed,
        ``False`` otherwise.
        """

        svalue   = self.__realToSlider(value)
        oldValue = self.GetValue()

        if svalue < self.__sliderMin: svalue = self.__sliderMin
        if svalue > self.__sliderMax: svalue = self.__sliderMax

        self.__slider.SetValue(svalue)
        self.__lastSliderValue = svalue

        return not np.isclose(value, oldValue)


    def GetValue(self):
        """Returns the slider value."""
        value = self.__slider.GetValue()
        return self.__sliderToReal(value)


class SliderSpinPanel(wx.Panel):
    """A panel containing a :class:`FloatSlider` and a :class:`.FloatSpinCtrl`.

    The slider and spinbox are linked such that changes to one are reflected
    in the other.  The ``SliderSpinPanel`` class also provides the option to
    have the minimum/maximum limits displayed on either side of the
    slider/spinbox, and to have those limits editable via a
    :class:`.NumberDialog` .

    Users of the ``SliderSpinPanel`` may wish to bind listeners to
    the following events:

      - :data:`EVT_SSP_VALUE`: Emitted when the slider value changes.

      - :data:`EVT_SSP_LIMIT`: Emitted when the slider limits change.
    """

    def __init__(self,
                 parent,
                 value=None,
                 minValue=None,
                 maxValue=None,
                 label=None,
                 style=None,
                 spinWidth=None):
        """Create a ``SliderSpinPanel``.

        The following style flags are available:

         .. autosummary::
            SSP_SHOW_LIMITS
            SSP_EDIT_LIMITS
            SSP_MOUSEWHEEL
            SSP_INTEGER

        :arg parent:   A parent control.

        :arg value:    Initial slider/spin value.

        :arg minValue: Minimum slider/spin value.

        :arg maxValue: Maximum slider/spin value.

        :arg label:    If not ``None``, a :class:`wx.StaticText`
                       widget is added to the left of the slider,
                       containing the given label.

        :arg style:    A combination of :data:`SSP_SHOW_LIMITS`,
                       :data:`SSP_EDIT_LIMITS`, :data:`SSP_NO_LIMITS`,
                       :data:`SSP_MOUSEWHEEL` and
                       :data:`SSP_INTEGER`. Defaults to
                       :data:`SSP_SHOW_LIMITS`.

        :arg spinWidth: Desired spin control width in characters. See the
                        :class:`.FloatSpinCtrl` class.
        """

        wx.Panel.__init__(self, parent, style=0)

        if style is None: style = SSP_SHOW_LIMITS

        showLimits = style & SSP_SHOW_LIMITS
        editLimits = style & SSP_EDIT_LIMITS
        noLimits   = style & SSP_NO_LIMITS
        integer    = style & SSP_INTEGER
        mousewheel = style & SSP_MOUSEWHEEL

        if value    is None: value    = 0
        if minValue is None: minValue = 0
        if maxValue is None: maxValue = 1

        if not showLimits:
            editLimits = False

        self.__showLimits = showLimits

        if integer: self.__fmt = '{}'
        else:       self.__fmt = '{: 0.3G}'

        spinStyle   = 0
        sliderStyle = 0

        if noLimits:
            spinStyle |= floatspin.FSC_NO_LIMIT

        if mousewheel:
            spinStyle   |= floatspin.FSC_MOUSEWHEEL
            sliderStyle |=            FS_MOUSEWHEEL

        if integer:
            spinStyle   |= floatspin.FSC_INTEGER
            sliderStyle |=            FS_INTEGER
            increment    = 1
        else:
            increment    = (maxValue - minValue) / 100.0

        self.__slider = FloatSlider(
            self,
            value=value,
            minValue=minValue,
            maxValue=maxValue,
            style=sliderStyle)

        self.__spinbox = floatspin.FloatSpinCtrl(
            self,
            value=value,
            minValue=minValue,
            maxValue=maxValue,
            increment=increment,
            style=spinStyle,
            width=spinWidth)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)

        if label is not None:
            self.__label = wx.StaticText(self, label=label)
            self.__sizer.Add(self.__label, flag=wx.EXPAND)

        self.__sizer.Add(self.__slider,  flag=wx.EXPAND, proportion=1)
        self.__sizer.Add(self.__spinbox, flag=wx.EXPAND)

        self.__slider .Bind(wx.EVT_SLIDER,           self.__onSlider)
        self.__spinbox.Bind(floatspin.EVT_FLOATSPIN, self.__onSpin)

        if showLimits:
            self.__minButton = wx.Button(self,
                                         label=self.__fmt.format(minValue))
            self.__maxButton = wx.Button(self,
                                         label=self.__fmt.format(maxValue))

            self.__sizer.Insert(0, self.__minButton, flag=wx.EXPAND)
            self.__sizer.Add(      self.__maxButton, flag=wx.EXPAND)

            self.__minButton.Enable(editLimits)
            self.__maxButton.Enable(editLimits)

            self.__minButton.Bind(wx.EVT_BUTTON, self.__onLimitButton)
            self.__maxButton.Bind(wx.EVT_BUTTON, self.__onLimitButton)

        self.Layout()


    @property
    def slider(self):
        """Returts a reference to the ``FloatSlider`` contained in this
        ``SliderSpinPanel``.
        """
        return self.__slider


    @property
    def spinCtrl(self):
        """Returts a reference to the ``FloatSpinCtrl`` contained in this
        ``SliderSpinPanel``.
        """
        return self.__spinbox


    def __onLimitButton(self, ev):
        """Only called if the :data:`SSP_SHOW_LIMITS` and
        :data:`SSP_EDIT_LIMITS` style flags are set.

        Called when either of the minimum/maximum limit buttons are clicked.

        Pops up a :class:`.NumberDialog` window. If the user changes the limit
        value, updates the slider/spin limits, and emits an
        :data:`EVT_SSP_LIMIT` event.
        """

        source = ev.GetEventObject()

        if source == self.__minButton:
            message = 'New minimum value'
            initVal = self.GetMin()
            minVal  = None
            maxVal  = self.GetMax()
        elif source == self.__maxButton:
            message = 'New maximum value'
            initVal = self.GetMax()
            minVal  = self.GetMin()
            maxVal  = None
        else:
            return

        dlg = numberdialog.NumberDialog(
            self.GetTopLevelParent(),
            message=message,
            initial=initVal,
            minValue=minVal,
            maxValue=maxVal)

        pos = ev.GetEventObject().GetScreenPosition()
        dlg.SetPosition(pos)
        if dlg.ShowModal() != wx.ID_OK:
            return

        # The NumberDialog should not return
        # an invalid value, but just in case,
        # we absorb any ValueErrors that are
        # raised.
        try:
            if   source == self.__minButton: self.SetMin(dlg.GetValue())
            elif source == self.__maxButton: self.SetMax(dlg.GetValue())
        except ValueError:
            return

        wx.PostEvent(self, SliderSpinLimitEvent(
            min=self.GetMin(),
            max=self.GetMax()))


    def __onSlider(self, ev):
        """Called when the user changes the slider value.

        Updates the spinbox value and emits an :data:`EVT_SSP_VALUE` event.
        """
        val = self.__slider.GetValue()
        self.__spinbox.SetValue(val)
        wx.PostEvent(self, SliderSpinValueEvent(value=val))


    def __onSpin(self, ev):
        """Called when the user changes the spinbox value.

        Updates the slider value and emits an :data:`EVT_SSP_VALUE` event.
        """
        val = self.__spinbox.GetValue()
        self.__slider.SetValue(val)
        wx.PostEvent(self, SliderSpinValueEvent(value=val))


    def GetRange(self):
        """Return a tuple containing the (minimum, maximum) slider/spinbox
        values.
        """
        return self.__slider.GetRange()


    def GetMin(self):
        """Returns the minimum slider/spinbox value."""
        return self.__slider.GetMin()


    def GetMax(self):
        """Returns the maximum slider/spinbox value."""
        return self.__slider.GetMax()


    def GetValue(self):
        """Returns the current slider/spinbox value. """

        # If SSP_NO_LIMITS is set, the slider might
        # return a value limited to the min/max, but
        # the spinbox will return an unlimited value.
        return self.__spinbox.GetValue()


    def SetRange(self, minValue, maxValue):
        """Sets the minimum/maximum slider/spinbox values."""
        self.__slider .SetRange(minValue, maxValue)
        self.__spinbox.SetRange(minValue, maxValue)

        if self.__showLimits:
            self.__minButton.SetLabel(self.__fmt.format(minValue))
            self.__maxButton.SetLabel(self.__fmt.format(maxValue))


    def SetMin(self, minValue):
        """Sets the minimum slider/spinbox value."""
        self.SetRange(minValue, self.GetMax())


    def SetMax(self, maxValue):
        """Sets the maximum slider/spinbox value."""
        self.SetRange(self.GetMin(), maxValue)


    def SetValue(self, value):
        """Sets the current slider/spinbox value."""
        self.__slider .SetValue(value)
        self.__spinbox.SetValue(value)


FS_MOUSEWHEEL = 1
"""If set, mouse wheel events on the slider will change the value. """


FS_INTEGER = 2
"""If set, the slider will store an integer value instead of a floating point
value.
"""


_SliderSpinValueEvent, _EVT_SSP_VALUE = wxevent.NewEvent()
_SliderSpinLimitEvent, _EVT_SSP_LIMIT = wxevent.NewEvent()


EVT_SSP_VALUE        = _EVT_SSP_VALUE
"""Identifier for the :data:`SliderSpinValueEvent`."""


EVT_SSP_LIMIT        = _EVT_SSP_LIMIT
"""Identifier for the :data:`SliderSpinLimitEvent`."""


SliderSpinValueEvent = _SliderSpinValueEvent
"""Event emitted when the :class:`SliderSpinPanel` value
changes. Contains a single attribute, ``value``, which
contains the new value.
"""


SliderSpinLimitEvent = _SliderSpinLimitEvent
"""Event emitted when the :class:`SliderSpinPanel` limits
change. Contains two attributes, ``min`` and ``max``, which
contain the new limit values.
"""


SSP_SHOW_LIMITS = 1
"""If set, the data range limits are shown alongside the slider/spin control
widgets.
"""


SSP_EDIT_LIMITS = 2
"""If set, and :data:`SSP_SHOW_LIMITS` is also set, the data range limits
are shown on buttons alongside the slider/spin control widgets.

When the user pushes the button, a :class:`.NumberDialog` is shown, allowing
the user to change the data range.
"""

SSP_MOUSEWHEEL = 4
"""If set, mouse wheel events on the slider/spin controls will change the
value.
"""


SSP_INTEGER = 8
"""If set, the ``SliderSpinPanel`` will store an integer value instead of
floating point value.
"""


SSP_NO_LIMITS = 16
"""If set, the user is able to enter values outside of the range into the
spin controls.
"""



#!/usr/bin/env python
#
# test_colourbarbitmap.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path   as op
import itertools as it

import matplotlib.image as mplimg

import pytest

import fsleyes_widgets.utils.colourbarbitmap as cbarbmp

from . import compare_images


datadir = op.join(op.dirname(__file__), 'testdata', 'colourbarbitmap')


def _compare(bmp, fname):

    bmp       = bmp.transpose((1, 0, 2))
    fname     = op.join(datadir, fname)
    benchmark = mplimg.imread(fname) * 255

    result = compare_images(bmp, benchmark, 1)

    if not result[0]:
        print(result)

    return result[0]


def test_standard_usage():

    cmaps           = ['Greys', 'Blues']
    sizes           = [(100, 25)]
    cmapResolutions = [6, 256]
    orientations    = ['vertical', 'horizontal']
    alphas          = [0.25, 1.0]
    bgColours       = [(0, 0, 0, 0), (0, 0, 0, 1), (1, 0, 0, 1)]

    testcases = it.product(cmaps,
                           sizes,
                           cmapResolutions,
                           orientations,
                           alphas,
                           bgColours)

    for testcase in testcases:
        cmap, size, cmapRes, orient, alpha, bgColour = testcase

        if orient == 'vertical': height, width = size
        else:                    width, height = size

        bmp = cbarbmp.colourBarBitmap(cmap,
                                      width,
                                      height,
                                      cmapResolution=cmapRes,
                                      alpha=alpha,
                                      orientation=orient,
                                      bgColour=bgColour)

        fname = '_'.join(map(str, [
            cmap,
            cmapRes,
            orient,
            alpha] + list(bgColour)))

        fname = 'standard_usage_{}.png'.format(fname)

        assert _compare(bmp, fname)


def test_negCmap_invert():

    # cmap, negCmap, invert, orientation
    testcases = [
        ('Reds',  None,   False, 'vertical'),
        ('Reds',  None,   False, 'horizontal'),
        ('Reds',  None,   True,  'vertical'),
        ('Reds',  None,   True,  'horizontal'),
        ('Reds', 'Blues', False, 'vertical'),
        ('Reds', 'Blues', False, 'horizontal'),
        ('Reds', 'Blues', True,  'vertical'),
        ('Reds', 'Blues', True,  'horizontal')
    ]

    size = 100, 25

    for cmap, negCmap, invert, orient in testcases:

        fname = '_'.join(map(str, [cmap, negCmap, invert, orient]))
        fname = 'negCmap_invert_{}.png'.format(fname)

        if orient == 'vertical': height, width = size
        else:                    width, height = size

        bmp = cbarbmp.colourBarBitmap(cmap,
                                      width,
                                      height,
                                      negCmap=negCmap,
                                      invert=invert,
                                      orientation=orient)

        assert _compare(bmp, fname)


def test_label():

    orientations = ['horizontal', 'vertical']
    labelsides   = ['top', 'bottom', 'left', 'right']
    fontsizes    = [6, 10, 16]
    textcolours  = [(0, 0, 0, 1), (1, 0, 0, 1)]

    testcases = it.product(orientations, labelsides, fontsizes, textcolours)

    for orient, side, size, colour in testcases:

        if orient[0] == 'v': width, height = 100, 300
        else:                width, height = 300, 100

        bmp = cbarbmp.colourBarBitmap('Reds',
                                      width,
                                      height,
                                      label='Label',
                                      orientation=orient,
                                      labelside=side,
                                      fontsize=size,
                                      textColour=colour)

        fname = '_'.join(map(str, [orient, side, size] + list(colour)))
        fname = 'label_{}.png'.format(fname)

        assert _compare(bmp, fname)


def test_ticks():
    orientations = ['horizontal', 'vertical']
    labelsides   = ['top', 'bottom', 'left', 'right']
    fontsizes    = [6, 10, 16]
    textcolours  = [(0, 0, 0, 1), (1, 0, 0, 1)]

    tickss      = [[0.0, 0.25, 0.75, 1.0],
                   [0.25, 0.75]]

    testcases = it.product(orientations,
                           labelsides,
                           fontsizes,
                           textcolours,
                           tickss)

    for orient, side, size, colour, ticks in testcases:

        labels = [str(t) for t in ticks]

        if orient[0] == 'v': width, height = 100, 300
        else:                width, height = 300, 100

        bmp = cbarbmp.colourBarBitmap('Reds',
                                      width,
                                      height,
                                      ticks=ticks,
                                      ticklabels=labels,
                                      orientation=orient,
                                      labelside=side,
                                      fontsize=size,
                                      textColour=colour)

        fname = [orient, side, size] + list(colour) + labels
        fname = '_'.join(map(str, fname))
        fname = 'ticks_{}.png'.format(fname)

        assert _compare(bmp, fname)


def test_tickalign():

    orientations = ['horizontal', 'vertical']
    labelsides   = ['top', 'bottom']

    ticks      = [0.0, 0.25, 0.5, 0.75, 1.0]
    ticklabels = [str(t) for t in ticks]
    tickaligns = [
        ['left'   for t in ticks],
        ['right'  for t in ticks],
        ['center' for t in ticks],
        ['left', 'center', 'center', 'center', 'right']]

    testcases = it.product(orientations, labelsides, tickaligns)

    for orient, side, align in testcases:

        if orient[0] == 'v': width, height = 100, 300
        else:                width, height = 300, 100

        bmp = cbarbmp.colourBarBitmap('Blues',
                                      width,
                                      height,
                                      ticks=ticks,
                                      ticklabels=ticklabels,
                                      tickalign=align,
                                      orientation=orient,
                                      labelside=side,
                                      textColour=(0, 0, 0, 1))


        falign = ''.join([a[0] for a in align])
        fname  = '_'.join([orient, side, falign])
        fname  = 'tickalign_{}.png'.format(fname)

        assert _compare(bmp, fname)


def test_label_and_ticks():

    orientations = ['horizontal', 'vertical']
    labelsides   = ['top', 'bottom']
    fontsizes    = [6, 10, 16]

    testcases = it.product(orientations,
                           labelsides,
                           fontsizes)

    ticks      = [0.2, 0.5, 0.8]
    ticklabels = [str(t) for t in ticks]
    label      = 'Label'

    for orient, side, size, in testcases:

        if orient[0] == 'v': width, height = 100, 300
        else:                width, height = 300, 100

        bmp = cbarbmp.colourBarBitmap('Reds',
                                      width,
                                      height,
                                      ticks=ticks,
                                      ticklabels=ticklabels,
                                      label=label,
                                      orientation=orient,
                                      labelside=side,
                                      fontsize=size,
                                      textColour=(0, 0, 0.2, 1))

        fname = '_'.join([orient, side, str(size)])
        fname = 'label_and_ticks_{}.png'.format(fname)

        assert _compare(bmp, fname)


def test_badargs():

    testcases = [
        dict(cmap='badcmap'),
        dict(orientation='badorient'),
        dict(labelside='badside'),
        dict(ticks=[0, 0.5, 1],
             ticklabels=['l', 'l', 'l'],
             tickalign=['badalign', 'badalign', 'badalign']),
    ]

    # bad colour map
    for t in testcases:
        with pytest.raises(Exception):
            cbarbmp.colourBarBitmap(t.get('cmap', 'Reds'), 50, 50, **t)



#!/usr/bin/env python
#
# textpanel.py - A panel for displaying horizontal or vertical text.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TextPanel` class, for displaying
some text, oriented either horizontally or vertically.
"""


import wx

import fsleyes_widgets as fw


if fw.wxversion() == fw.WX_PHOENIX: TextPanelBase = wx.Panel
else:                               TextPanelBase = wx.PyPanel


class TextPanel(TextPanelBase):
    """A :class:`wx.PyPanel` which may be used to display a string of
    text, oriented either horizotnally or vertically.
    """

    def __init__(self, parent, text=None, orient=wx.HORIZONTAL):
        """Create a ``TextPanel``.

        :arg parent: The :mod:`wx` parent object.

        :arg text:   The text to display. This can be changed via
                     :meth:`SetText`.

        :arg orient: Text orientation - either ``wx.HORIZONTAL`` (the
                     default) or ``wx.VERTICAL``. This can be changed
                     later via :meth:`SetOrient`.
        """
        TextPanelBase.__init__(self, parent)

        self.Bind(wx.EVT_PAINT, self.Draw)
        self.Bind(wx.EVT_SIZE,  self.__onSize)

        self.__text   = text
        self.__size   = None
        self.__orient = None

        self.SetOrient(orient)


    def SetOrient(self, orient):
        """Sets the orientatino of the text on this ``TextPanel``.

        :arg orient: Either ``wx.HORIZONTAL`` or ``wx.VERTICAL``.
        """

        if orient not in (wx.HORIZONTAL, wx.VERTICAL):
            raise ValueError('TextPanel orient must be '
                             'wx.HORIZONTAL or wx.VERTICAL')

        self.__orient = orient

        # trigger re-calculation of
        # text extents and a refresh
        self.SetLabel(self.__text)


    def DoGetBestClientSize(self):
        """Returns the best (minimum) size for this ``TextPanel``. """

        size = wx.Size(self.__size)
        self.CacheBestSize(size)
        return size


    def SetLabel(self, text):
        """Sets the text shown on this ``TextPanel``."""

        dc = wx.ClientDC(self)

        self.__text = text

        if text is None:
            self.SetMinSize((0, 0))
            return

        width, height = dc.GetTextExtent(text)

        if self.__orient == wx.VERTICAL:
            width, height = height, width

        self.__size = (width, height)

        self.SetMinSize((width, height))

        self.Refresh()


    def __onSize(self, ev):
        """Called when this ``TextPanel`` is resized. Triggers a refresh. """
        self.Refresh()
        ev.Skip()


    def Draw(self, ev=None):
        """Draws this ``TextPanel``. """

        self.ClearBackground()

        if self.__text is None or self.__text == '':
            return

        if ev is None: dc = wx.ClientDC(self)
        else:          dc = wx.PaintDC( self)

        if not dc.IsOk():
            return

        paneW, paneH = dc.GetSize().Get()
        textW, textH = self.__size

        x = (paneW - textW) / 2.0
        y = (paneH - textH) / 2.0

        if self.__orient == wx.VERTICAL:
            dc.DrawRotatedText(self.__text, x, paneH - y, 90)
        else:
            dc.DrawText(self.__text, x, y)



#!/usr/bin/env python
#
# layout.py - Utility functions for calculating canvas sizes and laying them
# out.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Utility functions for calculating canvas sizes and laying them out.


This module implements a simple layout manager, for laying out canvases and
associated orientation labels. It is used primarily by the :mod:`.render`
application, for off-screen rendering.

You can use the following classes to define a layout:

.. autosummary::
   :nosignatures:

   Bitmap
   Space
   HBox
   VBox


And the following functions to generate layouts and bitmaps:

.. autosummary::
   :nosignatures:

   buildOrthoLayout
   buildCanvasBox
   padBitmap
   layoutToBitmap


A few functions are also provided for calculating the display size, in pixels,
of one or more canvases which are displaying a defined coordinate system. The
canvas sizes are calculated so that their aspect ratio, relative to the
respective horizontal/vertical display axes, are maintained, and that the
canvases are sized proportionally with respect to each other. The following
size calculation functions are available:

.. autosummary::
   :nosignatures:

   calcSizes
   calcGridSizes
   calcHorizontalSizes
   calcVerticalSizes
   calcPixWidth
   calcPixHeight
"""


import logging

import numpy as np


log = logging.getLogger(__name__)


class Bitmap(object):
    """A class which encapsulates a RGBA bitmap, assumed to be a
    ``numpy.uint8`` array of shape :math:`height \\times width \\times 4`).

    .. warning:: Note the unusual array shape - height is the first axis,
                 and width the second!


    A ``Bitmap`` instance has the following attributes:

      - ``bitmap``: The bitmap data
      - ``width``:  Bitmap width in pixels
      - ``height``: Bitmap height in pixels
    """

    def __init__(self, bitmap):
        """Create a ``Bitmap``.

        :arg bitmap: :mod:`numpy` array containing the bitmap data.
        """
        self.bitmap = bitmap
        self.width  = bitmap.shape[1]
        self.height = bitmap.shape[0]


class Space(object):
    """A class which represents empty space of a specific width/height.


    A ``Space`` instance has the following attributes:

      - ``width``:  Width in pixels.
      - ``height``: Height in pixels.
    """

    def __init__(self, width, height):
        """Creat a ``Space``.

        :arg width:  Width in pixels.

        :arg height: Height in pixels.
        """
        self.width  = width
        self.height = height


class HBox(object):
    """A class which contains items to be laid out horizontally.

    After creation, new items should be added via the :meth:`append` method.

    A ``HBox`` instance has the following attributes:

      - ``width``:  Total width in pixels.
      - ``height``: Total height in pixels.
      - ``items``:  List of items in this ``HBox``.
    """


    def __init__(self, items=None):
        """Create a ``HBox``.

        :arg items: List of items contained in this ``HBox``.
        """
        self.width  = 0
        self.height = 0
        self.items  = []
        if items is not None:
            for i in items:
                self.append(i)

    def append(self, item):
        """Append a new item to this ``HBox``. """
        self.items.append(item)
        self.width = self.width + item.width
        if item.height > self.height:
            self.height = item.height


class VBox(object):
    """A class which contains items to be laid out vertically.

    After creation, new items can be added via the :meth:`append` method.

    A ``VBox`` instance has the following attributes:

      - ``width``:  Total width in pixels.
      - ``height``: Total height in pixels.
      - ``items``:  List of items in this ``VBox``.
    """

    def __init__(self, items=None):
        """Create a ``VBox``.

        :arg items: List of items contained in this ``VBox``.
        """
        self.width  = 0
        self.height = 0
        self.items  = []
        if items is not None:
            for i in items:
                self.append(i)


    def append(self, item):
        """Append a new item to this ``VBox``. """
        self.items.append(item)
        self.height = self.height + item.height
        if item.width > self.width:
            self.width = item.width


def padBitmap(bitmap, width, height, vert, bgColour):
    """Pads the given bitmap with zeros along the secondary axis (specified
    with the ``vert`` parameter), so that it fits in the given
    ``width``/``height``.


    :arg bitmap:   A ``numpy.array`` of size :math:`x \\times y \\times 4`
                   containing a RGBA bitmap.

    :arg width:    Desired width in pixels.

    :arg height:   Desired height in pixels.

    :arg vert:     If ``vert`` is ``True``, the bitmap is padded
                   horizontally to fit ``width``. Otherwise, the
                   bitmap is padded vertically to fit ``height``.

    :arg bgColour: Background colour to use for padding. Must be
                   a ``(r, g, b, a)`` tuple with each channel in
                   the range ``[0 - 255]``.
    """

    iheight = bitmap.shape[0]
    iwidth  = bitmap.shape[1]

    if vert:
        if iwidth < width:
            lpad   = int(np.floor((width - iwidth) / 2.0))
            rpad   = int(np.ceil( (width - iwidth) / 2.0))
            lpad   = np.zeros((iheight, lpad, 4), dtype=np.uint8)
            rpad   = np.zeros((iheight, rpad, 4), dtype=np.uint8)
            lpad[:] = bgColour
            rpad[:] = bgColour
            bitmap = np.hstack((lpad, bitmap, rpad))
    else:
        if iheight < height:
            tpad   = int(np.floor((height - iheight) / 2.0))
            bpad   = int(np.ceil(( height - iheight) / 2.0))
            tpad   = np.zeros((tpad, iwidth, 4), dtype=np.uint8)
            bpad   = np.zeros((bpad, iwidth, 4), dtype=np.uint8)
            tpad[:] = bgColour
            bpad[:] = bgColour
            bitmap = np.vstack((tpad, bitmap, bpad))

    return bitmap


def layoutToBitmap(layout, bgColour=None):
    """Recursively turns the given ``layout`` object into a bitmap.

    :arg layout:   A :class:`Bitmap`, :class:`Space`, :class:`HBox` or
                   :class:`VBox` instance.

    :arg bgColour: Background colour used to fill in empty space. Must be
                   a ``(r, g, b, a)`` tuple with channel values in the range
                   ``[0, 255]``. Defaults to transparent.

    :returns:      a ``numpy.uint8`` array of size
                   :math:`height \\times width \\times 4`.
    """

    if bgColour is None: bgColour = [0, 0, 0, 0]
    bgColour = np.array(bgColour, dtype=np.uint8)

    # Space is easy
    if isinstance(layout, Space):
        space = np.zeros((layout.height, layout.width, 4), dtype=np.uint8)
        space[:] = bgColour
        return space

    # Bitmap is easy
    elif isinstance(layout, Bitmap):
        return np.array(layout.bitmap, dtype=np.uint8)

    # Boxes require a bit of work
    if   isinstance(layout, HBox): vert = False
    elif isinstance(layout, VBox): vert = True

    # Recursively bitmapify the children of the box
    itemBmps = [layoutToBitmap(i, bgColour) for i in layout.items]

    # Pad each of the bitmaps so they are all the same
    # size along the secondary axis (which is width
    # if the layout is a VBox, and height if the layout
    # is a HBox).
    width    = layout.width
    height   = layout.height
    itemBmps = [padBitmap(i, width, height, vert, bgColour) for i in itemBmps]

    if len(itemBmps) == 0:
        bmp         = np.zeros((height, width, 4))
        bmp[..., :] = bgColour
        return bmp

    if vert: return np.vstack(itemBmps)
    else:    return np.hstack(itemBmps)


def buildCanvasBox(canvasBmp, labelBmps, showLabels, labelSize):
    """Builds a layout containing the given canvas bitmap, and orientation
    labels (if ``showLabels`` is ``True``).


    :arg canvasBmp:  A ``numpy.uint8`` array containing a bitmap.

    :arg labelBmps:  Only used if ``showLabels`` is ``True``. ``numpy.uint8``
                     arrays containing label bitmaps. Must be a
                     dictionary of ``{side : numpy.uint8}`` mappings,
                     and must have keys ``top``, ``bottom``, ``left`` and
                     ``right``.

    :arg showLabels: If ``True``, the orientation labels provided in
                     ``labelBmps`` are added to the layout.

    :arg labelSize:  Label sizes - the ``left``/``right`` label widths,
                     and ``top``/``bottom`` label heights are padded to this
                     size using ``Space`` objects.

    :returns:        A :class:`Bitmap`  or :class:`VBox` instance.
    """

    if not showLabels: return Bitmap(canvasBmp)

    row1Box = HBox([Space(labelSize, labelSize),
                    Bitmap(labelBmps['top']),
                    Space(labelSize, labelSize)])

    row2Box = HBox([Bitmap(labelBmps['left']),
                    Bitmap(canvasBmp),
                    Bitmap(labelBmps['right'])])

    row3Box = HBox([Space(labelSize, labelSize),
                    Bitmap(labelBmps['bottom']),
                    Space(labelSize, labelSize)])

    return VBox((row1Box, row2Box, row3Box))


def buildOrthoLayout(canvasBmps,
                     labelBmps,
                     layout,
                     showLabels,
                     labelSize):
    """Builds a layout containing the given canvas bitmaps, label bitmaps, and
    colour bar bitmap.


    :arg canvasBmps: A list of ``numpy.uint8`` arrays containing the canvas
                     bitmaps to be laid out.

    :arg layout:     One of ``'horizontal'``, ``'vertical'``, or ``'grid'``.

    See the :func:`buildCanvasBox` for details on the other parameters.


    :returns: A :class:`HBox` or :class:`VBox` describing the layout.
    """

    if labelBmps is None:
        labelBmps  = [None] * len(canvasBmps)
        showLabels = False

    canvasBoxes = [buildCanvasBox(cbmp, lbmps, showLabels, labelSize)
                   for cbmp, lbmps in zip(canvasBmps, labelBmps)]

    if   layout == 'horizontal': canvasBox = HBox(canvasBoxes)
    elif layout == 'vertical':   canvasBox = VBox(canvasBoxes)
    elif layout == 'grid':
        row1Box   = HBox([canvasBoxes[0], canvasBoxes[1]])
        row2Box   = HBox([canvasBoxes[2], Space(canvasBoxes[1].width,
                                                canvasBoxes[2].height)])
        canvasBox = VBox((row1Box, row2Box))

    return canvasBox


def calcSizes(layout, canvasaxes, bounds, width, height):
    """Convenience function which, based upon whether the `layout` argument
    is ``'horizontal'``, ``'vertical'``, or ``'grid'``,  respectively calls
    one of:

      - :func:`calcHorizontalSizes`
      - :func:`calcVerticalSizes`
      - :func:`calcGridSizes`

    :arg layout:    String specifying the layout type.

    :arg canvsaxes: A list of tuples, one for each canvas to be laid out.
                    Each tuple contains two values, ``(i, j)``, where ``i``
                    is an index, into ``bounds``, specifying the canvas
                    width, and ``j`` is an index into ``bounds``, specifying
                    the canvas height, in the display coordinate system.

    :arg bounds:    A list of three values specifying the size of the display
                    space.

    :arg width:     Maximum width in pixels.

    :arg height:    Maximum height in pixels.

    :returns:       A list of ``(width, height)`` tuples, one for each canvas,
                    each specifying the canvas width and height in pixels.
    """

    layout = layout.lower()
    func   = None

    if   layout == 'horizontal': func = calcHorizontalSizes
    elif layout == 'vertical':   func = calcVerticalSizes
    elif layout == 'grid':       func = calcGridSizes

    # a bad value for layout
    # will result in an error
    sizes = func(canvasaxes, bounds, width, height)

    log.debug('For space ({}, {}) and {} layout, pixel '
              'sizes for canvases {} ({}) are: {}'.format(
                  width, height, layout, canvasaxes, bounds, sizes))

    return sizes


def calcGridSizes(canvasaxes, bounds, width, height):
    """Calculates the size of three canvases so that they are laid
    out in a grid, i.e.:

       0   1

       2


    It is assumed that the ``canvasaxes`` parameters are ordered such that
    adjacent canvases share common axes in the world coordinate system. In
    other words, in the above diagram, the horizontal axes of canvases ``0``
    and ``2`` are equal, and the vertical axes of canvases ``0`` and ``1``
    must be equal.


    .. note:: If less than three canvases are specified, they are passed to
              the :func:`calcHorizontalLayout` function.


    See :func:`calcSizes` for details on the arguments.
    """

    if len(canvasaxes) < 3:
        return calcHorizontalSizes(canvasaxes, bounds, width, height)

    # Warn if axes are not aligned
    if any((canvasaxes[0][1] != canvasaxes[1][1],
            canvasaxes[0][0] != canvasaxes[2][0])):
        log.warning('Canvas axes are not aligned: {}'.format(canvasaxes))

    canvasWidths  = [bounds[c[0]] for c in canvasaxes]
    canvasHeights = [bounds[c[1]] for c in canvasaxes]

    ttlWidth      = float(canvasWidths[ 0] + canvasWidths[ 1])
    ttlHeight     = float(canvasHeights[0] + canvasHeights[2])

    sizes = []

    for i in range(len(canvasaxes)):

        cw = width  * (canvasWidths[ i] / ttlWidth)
        ch = height * (canvasHeights[i] / ttlHeight)

        acw, ach = _adjustPixelSize(canvasWidths[ i],
                                    canvasHeights[i],
                                    cw,
                                    ch)

        if (float(cw) / ch) > (float(acw) / ach): cw, ch = cw, ach
        else:                                     cw, ch = acw, ch

        sizes.append((cw, ch))

    return sizes


def calcVerticalSizes(canvasaxes, bounds, width, height):
    """Calculates the size of up to three canvases so they are laid out
    vertically.

    See :func:`calcSizes` for details on the arguments.
    """
    return _calcFlatSizes(canvasaxes, bounds, width, height, True)


def calcHorizontalSizes(canvasaxes, bounds, width, height):
    """Calculates the size of up to three canvases so they are laid out
    horizontally.

    See :func:`calcSizes` for details on the arguments.
    """
    return _calcFlatSizes(canvasaxes, bounds, width, height, False)


def _calcFlatSizes(canvasaxes, bounds, width, height, vert=True):
    """Used by :func:`calcVerticalSizes` and :func:`calcHorizontalSizes`.

    Calculates the width and height, in pixels, of each canvas.

    :arg vert: If ``True`` the sizes are calculated for a vertical layout;
               otherwise they are calculated for a horizontal layout.

    See :func:`calcSizes` for details on the other arguments.

    :returns:  A list of ``(width, height)`` tuples, one for each canvas,
               each specifying the canvas width and height in pixels.
    """
    if len(canvasaxes) == 0:
        return []

    # Get the canvas dimensions in world space
    canvasWidths  = [bounds[c[0]] for c in canvasaxes]
    canvasHeights = [bounds[c[1]] for c in canvasaxes]

    maxWidth  = float(max(canvasWidths))
    maxHeight = float(max(canvasHeights))
    ttlWidth  = float(sum(canvasWidths))
    ttlHeight = float(sum(canvasHeights))

    if vert: ttlWidth  = maxWidth
    else:    ttlHeight = maxHeight

    sizes = []
    for i in range(len(canvasaxes)):

        if ttlWidth  == 0: cw = 0
        else:              cw = width  * (canvasWidths[ i] / ttlWidth)

        if ttlHeight == 0: ch = 0
        else:              ch = height * (canvasHeights[i] / ttlHeight)

        sizes.append((cw, ch))

    return sizes


def calcPixWidth(wldWidth, wldHeight, pixHeight):
    """Given the dimensions of a space to be displayed, and the available
    height in pixels, calculates the required pixel width.

    :arg wldWidth:   Width of the display coordinate system

    :arg wldHeight:  Height of the display coordinate system

    :arg pixHeight:  Available height in pixels.

    :returns:        The required width in pixels.
    """
    return _adjustPixelSize(wldWidth,
                            wldHeight,
                            pixHeight * (2 ** 32),
                            pixHeight)[0]


def calcPixHeight(wldWidth, wldHeight, pixWidth):
    """Given the dimensions of a space to be displayed, and the available
    width in pixels, calculates the required pixel height.

    :arg wldWidth:   Width of the display coordinate system

    :arg wldHeight:  Height of the display coordinate system

    :arg pixWidth:   Available width in pixels.

    :returns:        The required height in pixels.
    """
    return _adjustPixelSize(wldWidth,
                            wldHeight,
                            pixWidth,
                            pixWidth * (2 ** 32))[1]



def _adjustPixelSize(wldWidth, wldHeight, pixWidth, pixHeight):
    """Used by :func:`calcPixelWidth` and :func:`calcPixelHeight`.

    Potentially reduces the given pixel width/height such that the
    display space aspect ratio is maintained.
    """

    if any((pixWidth  == 0,
            pixHeight == 0,
            wldWidth  == 0,
            wldHeight == 0)):
        return 0, 0

    pixRatio = float(pixWidth) / pixHeight
    wldRatio = float(wldWidth) / wldHeight

    if   pixRatio > wldRatio:
        pixWidth  = wldWidth  * (pixHeight / wldHeight)

    elif pixRatio < wldRatio:
        pixHeight = wldHeight * (pixWidth  / wldWidth)

    return pixWidth, pixHeight



#!/usr/bin/env python
#
# textbitmap.py - A function which renders some text using matplotlib, and
# returns it as an RGBA bitmap.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a single function, :func:`textBitmap`, which renders
some text off-screen using :mod:`matplotlib`, and returns it as an RGBA bitmap.
"""


def textBitmap(text,
               width,
               height,
               fontSize,
               fgColour,
               bgColour,
               alpha=1.0):
    """Draw some text using :mod:`matplotlib`.


    The rendered text is returned as a RGBA bitmap within a ``numpy.uint8``
    array of size :math:`w \\times h \\times 4`, with the top-left pixel
    located at index ``[0, 0, :]``.

    :arg text:     Text to render.

    :arg width:    Width in pixels.

    :arg height:   Height in pixels.

    :arg fontSize: Font size in points.

    :arg fgColour: Foreground (text) colour - can be any colour specification
                   that is accepted by :mod:`matplotlib`.

    :arg bgColour: Background colour  - can be any colour specification that
                   is accepted by :mod:`matplotlib`..

    :arg alpha:    Text transparency, in the range ``[0.0 - 1.0]``.
    """

    # Imports are expensive
    import numpy                           as np
    import matplotlib.backends.backend_agg as mplagg
    import matplotlib.figure               as mplfig

    dpi    = 96.0
    fig    = mplfig.Figure(figsize=(width / dpi, height / dpi),
                           dpi=dpi)
    canvas = mplagg.FigureCanvasAgg(fig)
    ax     = fig.add_axes([0, 0, 1, 1])
    ax.axis('off')

    if bgColour is not None: fig.patch.set_facecolor(bgColour)
    else:                    fig.patch.set_alpha(0)

    ax.set_xticks([])
    ax.set_yticks([])

    ax.text(0.5,
            0.5,
            text,
            fontsize=fontSize,
            verticalalignment='center',
            horizontalalignment='center',
            transform=ax.transAxes,
            color=fgColour,
            alpha=alpha)

    fig.subplots_adjust(
        bottom=0,
        top=1,
        left=0,
        right=1)

    canvas.draw()
    buf = canvas.tostring_argb()
    ncols, nrows = canvas.get_width_height()

    bitmap = np.fromstring(buf, dtype=np.uint8)
    bitmap = bitmap.reshape(nrows, ncols, 4)

    rgb = bitmap[:, :, 1:]
    a   = bitmap[:, :, 0]
    bitmap = np.dstack((rgb, a))

    return bitmap



#!/usr/bin/env python
#
# test_typedict.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import itertools as it
import pytest

import fsleyes_widgets.utils.typedict as typedict


def test_create():

    td = typedict.TypeDict()


    assert len(td)          == 0
    assert len(td.keys())   == 0
    assert len(td.values()) == 0
    assert len(td.items())  == 0

    keys   = [0,
              1,
              2,
              'a',
              'b',
              'c',
              'a.b',
              'a.b.c',
              ('a', 'c'),
              ('a',  0)]
    values = list(range(len(keys)))

    td = typedict.TypeDict(        dict(zip(keys, values)))
    td = typedict.TypeDict(initial=dict(zip(keys, values)))

    print(     td)
    print(repr(td))

    tknkeys = [td.tokenifyKey(k) for k in keys]

    assert len(td)                       == len(keys)
    assert sorted(map(str, td.keys()))   == sorted(map(str, tknkeys))
    assert sorted(map(str, td.values())) == sorted(map(str, values))
    assert sorted(map(str, td.items()))  == sorted(map(str, zip(tknkeys, values)))

    for k, v in zip(keys, values):
        assert td[k] == v

    with pytest.raises(KeyError):
        td['non-existent']

    assert td.get('non-existent') is None
    assert td.get('non-existent', 'default') == 'default'


def test_class_keys():

    class A(object):
        pass

    class B(object):
        pass

    class C(object):
        pass

    td = typedict.TypeDict()

    a  = A()
    b  = B()
    c  = C()

    # TypeDict [currently] does not allow value
    # assignment with classes/instances as keys -
    # we must use the class name, encoded as a
    # string, when assigning.
    keycomponents = ['A', 'B', 'C', 'att1', 'att2', 'att3']
    keys          = list(it.chain(it.permutations(keycomponents, 1),
                                  it.permutations(keycomponents, 2),
                                  it.permutations(keycomponents, 3),
                                  it.permutations(keycomponents, 4),
                                  it.permutations(keycomponents, 5),
                                  it.permutations(keycomponents, 6)))

    for val, key in enumerate(keys):
        if len(key) == 1: td[key[0]] = val
        else:             td[key]    = val

    for val, key in enumerate(keys):

        # Keys can be passed as tuples
        assert td[key] == val

        # Or as dot-separated strings
        assert td['.'.join(key)] == val

        # But when accessing items,
        # we can use either class names,
        # classes, or instances.
        for toReplace in it.chain(it.combinations(['A', 'B', 'C'], 1),
                                  it.combinations(['A', 'B', 'C'], 2),
                                  it.combinations(['A', 'B', 'C'], 3)):

            clsrep  = {'A' : A, 'B' : B, 'C' : C}
            instrep = {'A' : a, 'B' : b, 'C' : c}

            # use class instead of name
            repkey = [clsrep[k] if k in toReplace else k for k in key]
            assert td[repkey] == val

            # use instance instead of name
            repkey = [instrep[k] if k in toReplace else k for k in key]
            assert td[repkey] == val


def test_class_hierarchy():

    class A(object):
        pass

    class B(A):
        pass

    td = typedict.TypeDict()
    td['A.a'] = 'A.a'
    td['A.b'] = 'A.b'
    td['A.c'] = 'A.c'
    td['A.d'] = 'A.d'
    td['B.a'] = 'B.a'
    td['B.b'] = 'B.b'
    td['B.1'] = 'B.1'
    td['B.2'] = 'B.2'

    a = A()
    b = B()

    assert td[A, 'a'] == 'A.a'
    assert td[A, 'b'] == 'A.b'
    assert td[A, 'c'] == 'A.c'
    assert td[A, 'd'] == 'A.d'

    assert td['A.a']  == 'A.a'
    assert td['A.b']  == 'A.b'
    assert td['A.c']  == 'A.c'
    assert td['A.d']  == 'A.d'

    assert td.get((A, 'a')) == 'A.a'
    assert td.get((A, 'b')) == 'A.b'
    assert td.get((A, 'c')) == 'A.c'
    assert td.get((A, 'd')) == 'A.d'

    assert td.get((A, 'a'), allhits=True) == ['A.a']
    assert td.get((A, 'b'), allhits=True) == ['A.b']
    assert td.get((A, 'c'), allhits=True) == ['A.c']
    assert td.get((A, 'd'), allhits=True) == ['A.d']

    assert td.get((a, 'a'), allhits=True) == ['A.a']
    assert td.get((a, 'b'), allhits=True) == ['A.b']
    assert td.get((a, 'c'), allhits=True) == ['A.c']
    assert td.get((a, 'd'), allhits=True) == ['A.d']

    assert td[B, 'a'] == 'B.a'
    assert td[B, 'b'] == 'B.b'
    assert td[B, '1'] == 'B.1'
    assert td[B, '2'] == 'B.2'

    assert td['B.a']  == 'B.a'
    assert td['B.b']  == 'B.b'
    assert td['B.1']  == 'B.1'
    assert td['B.2']  == 'B.2'

    assert td.get('B.a') == 'B.a'
    assert td.get('B.b') == 'B.b'
    assert td.get('B.1') == 'B.1'
    assert td.get('B.2') == 'B.2'

    assert td.get((B, 'a')) == 'B.a'
    assert td.get((B, 'b')) == 'B.b'
    assert td.get((B, '1')) == 'B.1'
    assert td.get((B, '2')) == 'B.2'
    assert td.get((b, 'a')) == 'B.a'
    assert td.get((b, 'b')) == 'B.b'
    assert td.get((b, '1')) == 'B.1'
    assert td.get((b, '2')) == 'B.2'

    with pytest.raises(KeyError):
        td['B.c']
    with pytest.raises(KeyError):
        td['B.d']

    assert td[B, 'c'] == 'A.c'
    assert td[B, 'd'] == 'A.d'
    assert td[b, 'c'] == 'A.c'
    assert td[b, 'd'] == 'A.d'

    assert td.get('B.a',    allhits=True) == ['B.a']
    assert td.get('B.b',    allhits=True) == ['B.b']
    assert td.get('B.1',    allhits=True) == ['B.1']
    assert td.get('B.2',    allhits=True) == ['B.2']
    assert td.get((B, 'a'), allhits=True) == ['B.a', 'A.a']
    assert td.get((B, 'b'), allhits=True) == ['B.b', 'A.b']
    assert td.get((B, '1'), allhits=True) == ['B.1']
    assert td.get((B, '2'), allhits=True) == ['B.2']
    assert td.get((B, 'c'), allhits=True) == ['A.c']
    assert td.get((B, 'd'), allhits=True) == ['A.d']

    assert td.get((b, 'a'), allhits=True) == ['B.a', 'A.a']
    assert td.get((b, 'b'), allhits=True) == ['B.b', 'A.b']
    assert td.get((b, '1'), allhits=True) == ['B.1']
    assert td.get((b, '2'), allhits=True) == ['B.2']
    assert td.get((b, 'c'), allhits=True) == ['A.c']

    assert td.get((B, 'a'), allhits=False, bykey=True) == 'B.a'
    assert td.get((B, 'b'), allhits=False, bykey=True) == 'B.b'
    assert td.get((B, '1'), allhits=False, bykey=True) == 'B.1'
    assert td.get((B, '2'), allhits=False, bykey=True) == 'B.2'

    assert td.get((B, 'a'), allhits=True,  bykey=True) == {('A', 'a') : 'A.a',
                                                           ('B', 'a') : 'B.a'}
    assert td.get((B, 'b'), allhits=True,  bykey=True) == {('A', 'b') : 'A.b',
                                                           ('B', 'b') : 'B.b'}
    assert td.get((B, '1'), allhits=True,  bykey=True) == {('B', '1') : 'B.1'}
    assert td.get((B, '2'), allhits=True,  bykey=True) == {('B', '2') : 'B.2'}
    assert td.get((B, 'c'), allhits=True,  bykey=True) == {('A', 'c') : 'A.c'}
    assert td.get((B, 'd'), allhits=True,  bykey=True) == {('A', 'd') : 'A.d'}

    assert td.get((b, 'a'), allhits=True,  bykey=True) == {('A', 'a') : 'A.a',
                                                           ('B', 'a') : 'B.a'}
    assert td.get((b, 'b'), allhits=True,  bykey=True) == {('A', 'b') : 'A.b',
                                                           ('B', 'b') : 'B.b'}
    assert td.get((b, '1'), allhits=True,  bykey=True) == {('B', '1') : 'B.1'}
    assert td.get((b, '2'), allhits=True,  bykey=True) == {('B', '2') : 'B.2'}
    assert td.get((b, 'c'), allhits=True,  bykey=True) == {('A', 'c') : 'A.c'}
    assert td.get((b, 'd'), allhits=True,  bykey=True) == {('A', 'd') : 'A.d'}

    assert td.get((A, 'a'), allhits=True,  bykey=True) == {('A', 'a') : 'A.a'}
    assert td.get((A, 'b'), allhits=True,  bykey=True) == {('A', 'b') : 'A.b'}
    assert td.get((A, 'c'), allhits=True,  bykey=True) == {('A', 'c') : 'A.c'}
    assert td.get((A, 'd'), allhits=True,  bykey=True) == {('A', 'd') : 'A.d'}



def test_exact():

    class A(object):
        pass

    class B(A):
        pass

    class C(A):
        pass

    td = typedict.TypeDict()

    td['A'] = 'A'
    td['B'] = 'B'

    assert td[C]     == 'A'
    assert td.get(C, allhits=True) == ['A']
    assert td.get(C) == 'A'
    assert td.get(C, allhits=True, exact=True) is None



#!/usr/bin/env python
#
# notebook.py - Re-implementation of the wx.Notebook widget
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

"""This module provides the :class:`Notebook` class, a notebook control
similar to the :class:`wx.Notebook`.
"""


import wx
import wx.lib.stattext as statictext


class Notebook(wx.Panel):
    """A :class:`wx.Panel` which provides :class:`wx.Notebook`-like
    functionality. Manages the display of multiple child windows. A row of
    buttons along the top allows the user to select which child window to
    display.

    This :class:`Notebook` implementation supports page enabling/disabling, and
    toggling of page visibility.
    """

    def __init__(self, parent):
        """Create a :class:`Notebook` object.

        :arg parent: The :mod:`wx` parent object.
        """

        wx.Panel.__init__(self, parent, style=wx.SUNKEN_BORDER)

        self.__buttonPanel = wx.Panel(self)

        self.__sizer       = wx.BoxSizer(wx.VERTICAL)
        self.__buttonSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.              SetSizer(self.__sizer)
        self.__buttonPanel.SetSizer(self.__buttonSizer)

        self.__dividerLine = wx.StaticLine(self, style=wx.LI_HORIZONTAL)

        # a row of buttons along the top
        self.__sizer.Add(
            self.__buttonPanel,
            border=5,
            flag=wx.EXPAND | wx.ALIGN_CENTER | wx.TOP | wx.RIGHT | wx.LEFT)

        # a horizontal line separating the buttons from the pages
        self.__sizer.Add(
            self.__dividerLine,
            border=5,
            flag=wx.EXPAND | wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT)

        # a vertical line at the start of the button row
        self.__buttonSizer.Insert(
            0,
            wx.StaticLine(self.__buttonPanel, style=wx.VERTICAL),
            border=3,
            flag=wx.EXPAND | wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT | wx.TOP)

        self.__pages    = []
        self.__buttons  = []
        self.__selected = None


    def __updateMinSize(self):
        """Calculate and return the best (minimum) size for this
        :class:`Notebook` instance.

        The returned size is the minimum size of the largest page, plus the
        size of the button panel.
        """

        buttonSize = self.__buttonPanel.GetBestSize()
        pageSizes  = [p.GetBestSize() for p in self.__pages]

        buttonWidth  = buttonSize[0]
        buttonHeight = buttonSize[1]

        divLineHeight = self.__dividerLine.GetMinSize()[0]

        pageWidths  = [ps[0] for ps in pageSizes]
        pageHeights = [ps[1] for ps in pageSizes]

        myWidth  = max([buttonWidth] + pageWidths)                 + 20
        myHeight = max(pageHeights) + buttonHeight + divLineHeight + 20

        self.SetMinSize((myWidth, myHeight))


    def FindPage(self, page):
        """Returns the index of the given page, or :data:`wx.NOT_FOUND`
        if the page is not in this notebook.
        """
        try:               return self.__pages.index(page)
        except ValueError: return wx.NOT_FOUND


    def InsertPage(self, index, page, text):
        """Inserts the given page into the notebook at the specified index. A
        button for the page is also added to the button row, with the specified
        text.
        """

        if (index > len(self.__pages)) or (index < 0):
            raise IndexError('Index out of range: {}'.format(index))

        # index * 2 because we add a vertical
        # line after every button (and + 1 for
        # the line at the start of the button row)
        button    = statictext.GenStaticText(self.__buttonPanel, label=text)
        buttonIdx = index * 2 + 1

        self.__pages.  insert(index, page)
        self.__buttons.insert(index, button)

        # index + 2 to account for the button panel and
        # the horizontal divider line (see __init__)
        self.__sizer.Insert(
            index + 2, page, border=5, flag=wx.EXPAND | wx.ALL, proportion=1)

        self.__buttonSizer.Insert(
            buttonIdx,
            button,
            flag=wx.ALIGN_CENTER)

        # A vertical line at the end of every button
        self.__buttonSizer.Insert(
            buttonIdx + 1,
            wx.StaticLine(self.__buttonPanel, style=wx.VERTICAL),
            border=3,
            flag=wx.EXPAND | wx.ALIGN_CENTER | wx.LEFT | wx.RIGHT | wx.TOP)

        # When the button is pushed, show the page
        # (unless the button has been disabled)
        def _showPage(ev):
            if not button.IsEnabled(): return
            self.SetSelection(self.FindPage(page))

        button.Bind(wx.EVT_LEFT_DOWN, _showPage)

        if self.__selected is None:
            self.__selected = 0
        self.SetSelection(self.__selected)

        page.Layout()
        self.__buttonPanel.Layout()
        self.Layout()

        self.__updateMinSize()


    def AddPage(self, page, text):
        """Adds the given page (and a corresponding button with the given
        text) to the end of the notebook.
        """
        self.InsertPage(len(self.__pages), page, text)


    def RemovePage(self, index):
        """Removes the page at the specified index, but does not destroy it.
        """

        if (index >= len(self.__pages)) or (index < 0):
            raise IndexError('Index out of range: {}'.format(index))

        buttonIdx = index * 2 + 1
        pageIdx   = index + 2

        self.__buttons.pop(index)
        self.__pages  .pop(index)

        # Destroy the button for this page (and the
        # vertical line that comes after the button)
        self.__buttonSizer.Remove(buttonIdx)
        self.__buttonSizer.Remove(buttonIdx + 1)

        # Remove the page but do not destroy it
        self.__pagePanel.Detach(pageIdx)

        if len(self.__pages) == 0:
            self.__selected = None

        self.__updateMinSize()


    def DeletePage(self, index):
        """Removes the page at the specified index, and (attempts to) destroy
        it.
        """
        page = self.__pages[index]
        self.RemovePage(index)
        page.Destroy()


    def GetSelection(self):
        """Returns the index of the currently selected page."""
        return self.__selected


    def SetSelection(self, index):
        """Sets the displayed page to the one at the specified index."""

        if index < 0 or index >= len(self.__pages):
            raise IndexError('Index out of range: {}'.format(index))

        self.__selected = index

        for i in range(len(self.__pages)):

            page     = self.__pages[  i]
            button   = self.__buttons[i]
            showThis = i == self.__selected

            if showThis:
                button.SetBackgroundColour('#ffffff')
                page.Show()
            else:
                button.SetBackgroundColour(None)
                page.Hide()

        button.Layout()
        self.__buttonPanel.Layout()
        self.Layout()
        self.Refresh()


    def AdvanceSelection(self, forward=True):
        """Selects the next (or previous, if ``forward``
        is ``False``) enabled page.
        """

        if forward: offset =  1
        else:       offset = -1

        newSelection = (self.GetSelection() + offset) % len(self.__pages)

        while newSelection != self.__selected:

            if self.__buttons[newSelection].IsEnabled():
                break

            newSelection = (self.__selected + offset) % len(self.__pages)

        self.SetSelection(newSelection)


    def EnablePage(self, index):
        """Enables the page at the specified index."""
        self.__buttons[index].Enable()


    def DisablePage(self, index):
        """Disables the page at the specified index."""
        self.__buttons[index].Disable()

        if self.GetSelection() == index:
            self.AdvanceSelection()

        self.Refresh()


    def ShowPage(self, index):
        """Shows the page at the specified index."""
        self.EnablePage(index)
        self.__buttons[index].Show()
        self.__pages[  index].Show()
        self.__buttonPanel.Layout()
        self.Refresh()


    def HidePage(self, index):
        """Hides the page at the specified index."""

        self.__buttons[index].Hide()
        self.__pages[  index].Hide()

        # we disable the page as well as hiding it,, as the
        # AdvanceSelection method, and button handlers, use
        # button.IsEnabled to determine whether a page is
        # active or not.
        self.DisablePage(index)

        self.__buttonPanel.Layout()
        self.__buttonPanel.Refresh()
